/********************
 * IrisCores.com 2023
 ********************/

#include <iostream>
#include <stdio.h>
#include <string>
#include <complex>
#include <math.h>
#include <chrono>

#include <opencv2/imgproc.hpp>
#include <opencv2/imgcodecs.hpp>

#include <boost/asio/io_service.hpp>
#include <boost/bind.hpp>
#include <boost/thread/thread.hpp>

using namespace cv;

/************************************************
 * Canvas Definitions (Position in Complex Plane) 
 ***********************************************/
const double X_Min = -2.0;
const double X_Max = +2.0;
const double Y_Min = -2.0;
const double Y_Max = +2.0;


/**************************************
 * @brief: Main Julia Computing Routine
 **************************************/
int julia_compute(std::complex<double> Plane_Point, std::complex<double>Julia ,int maxiter){
	std::complex < double > z = Plane_Point;
	for (int i = 0; i < maxiter; i++) {
		double abs_val = sqrt(real(z) * real(z) + imag(z) * imag(z));
		if (abs_val > 4.0) {
			return (i);
		}

		z = z * z + Julia;
	}

	return 0;
}

/******************************************
 *@brief : Main thread-wise Julia Rendering
 *****************************************/
void julia_render(uint8_t *Data, int x_dim ,int y_dim,int x_start,int x_stop,int y_start,int y_stop,std::complex<double> Julia,int Max_Iterations) {
	double X_Rend = X_Max - X_Min;
	double Y_Rend = Y_Max - Y_Min;

	for (int y = y_start; y < y_stop; y++) {
		for (int x = x_start; x < x_stop; x++) {

			// -------------------------------------
			// ------- Complex Plane Location ------
			double Point_Real = X_Min + double(x) / double(x_dim) * X_Rend;
			double Point_Imag = Y_Min + double(y) / double(y_dim) * Y_Rend;
			std::complex<double> Plane_Point = std::complex<double>(Point_Real, Point_Imag);

			// ---------------------------------
			// ------ Julia Point Compute ------
			int Res = julia_compute(Plane_Point, Julia, Max_Iterations);

			// ------------------------
			// ------- Save Data ------
			int Index = x + y * x_dim;
			Data[Index] = static_cast<uint8_t>(Res);

		}
	}
}


int main(int argc, char** argv) {
	int Num_Threads = 20;
	int Resolution = 4;
	int Max_Iter = 255;
	std::complex<double> Julia_Value = std::complex<double>(-0.4, 0.6);
	size_t Dims_X[5] = { 2560, 3840, 7680, 15360, 30720 };
	size_t Dims_Y[5] = { 1440, 2160, 4320, 8640, 17280 };
	size_t Image_Width = Dims_X[Resolution];
	size_t Image_Height = Dims_Y[Resolution];

	uint8_t* Julia_1C = new uint8_t[Image_Width * Image_Height * size_t(1)];
	uint8_t* Julia_4C = new uint8_t[Image_Width * Image_Height * size_t(4)];

	
	/************************************************************************
	* Create an asio::io_service and a thread_group (through pool in essence)
	**************************************************************************/
	boost::asio::io_service ioService;
	boost::thread_group threadpool;

	/**********************************************************
	 * This will start the ioService processing loop. All tasks
	 * assigned with ioService.post() will start executing.
	 **********************************************************/
	boost::asio::io_service::work work(ioService);

	// ----------------------------------------------------
	// ------Allocate Threads within the Thread Pool ------
	for (auto i = 0; i < Num_Threads; i++) {
		threadpool.create_thread(boost::bind(&boost::asio::io_service::run, &ioService));
	}

	// ---------------------------
	// ------- Start Timing ------
	auto start = std::chrono::high_resolution_clock::now();

	// ----------------------------
	// ------ Compute Julia -------
	int y_chunk = Image_Height / Num_Threads;
	for (auto i = 0; i < Num_Threads; i++) {
		int x_start = 0;
		int x_stop = Image_Width;
		int y_start = i * y_chunk;
		int y_stop = (i + 1) * y_chunk;
		ioService.post(boost::bind(julia_render, Julia_1C, Image_Width, Image_Height, x_start, x_stop, y_start, y_stop, Julia_Value, Max_Iter));
	}

	/************************************************************************
	 * Will wait till all the threads in the thread pool are finished with
	 * their assigned tasks and 'join' them. Just assume the threads inside
	 * the threadpool will be destroyed by this method.
	 *************************************************************************/
	ioService.stop();
	threadpool.join_all();
	
	auto stop = std::chrono::high_resolution_clock::now();

	// --------------------------------------------
	// ------ Convert with a custom Colormap ------
	cv::Mat Julia_CV_1C(cv::Size(Image_Width, Image_Height), CV_8UC1, Julia_1C);
	cv::Mat Julia_CV_4C(cv::Size(Image_Width, Image_Height), CV_8UC4, Julia_4C);
	applyColorMap(Julia_CV_1C, Julia_CV_4C, cv::COLORMAP_MAGMA);

	// ----------------------------------
	// ------ Write Image to File -------
	imwrite("Julia.png", Julia_CV_4C);

	delete[]Julia_1C;
	delete[]Julia_4C;

	auto duration = std::chrono::duration_cast<std::chrono::microseconds>(stop - start);
	std::cout << "Julia Compute Duration: " << duration.count() << " us / " << duration.count() / 1000 << " ms" << std::endl;
	system("pause");
	return 0;
}