/********************
 * IrisCores.com 2023
 ********************/
package main

//"context" // Context Signalling
import (
	"fmt" // Printline Stuff
	"image"
	"image/color"
	"image/png"
	"math" // Mathematical
	"os"
	"time"

	"github.com/mazznoer/colorgrad"

	// OS Functions
	"runtime" // go runtime configuration
	"sync"    // Thread Synchronization
)

/*******************************
 *@brief: Define Rendering Plane
 ******************************/
const X_Min float64 = -2.0
const X_Max float64 = +2.0
const Y_Min float64 = -2.0
const Y_Max float64 = +2.0

/**************************************
 * @brief: Main Julia Computing Routine
 **************************************/
func julia_compute(Plane_Point complex128, Julia complex128, maxiter int) int {
	var z complex128 = Plane_Point
	for i := 0; i < maxiter; i++ {
		var abs_val float64 = math.Sqrt(real(z)*real(z) + imag(z)*imag(z))
		if abs_val > 4.0 {
			return (i)
		}
		z = z*z + Julia
	}
	return 0
}

/******************************************
 *@brief : Main thread-wise Julia Rendering
 *****************************************/
func julia_render(wg *sync.WaitGroup, Data *image.NRGBA, x_dim int, y_dim int, x_start int, x_stop int, y_start int, y_stop int, Julia complex128, Max_Iterations int, ThreadID int) {
	X_Rend := X_Max - X_Min
	Y_Rend := Y_Max - Y_Min

	grad := colorgrad.Magma()
	for y := y_start; y < y_stop; y++ {
		for x := x_start; x < x_stop; x++ {

			// -------------------------------------
			// ------- Complex Plane Location ------
			Point_Real := X_Min + float64(x)/float64(x_dim)*X_Rend
			Point_Imag := Y_Min + float64(y)/float64(y_dim)*Y_Rend
			Plane_Point := complex(Point_Real, Point_Imag)

			// ---------------------------------
			// ------ Julia Point Compute ------
			Res := julia_compute(Plane_Point, Julia, Max_Iterations)

			// ---------------------------
			// ------ Color Mapping ------
			Pix := grad.At(float64(Res) / 255.0)

			// ------------------------
			// ------- Save Data ------
			Data.Set(x, y, color.NRGBA{
				R: uint8(Pix.R * 255.0),
				G: uint8(Pix.G * 255.0),
				B: uint8(Pix.B * 255.0),
				A: 255,
			})

		}
	}
	defer wg.Done()
}

/**************************
 *@brief: Julia-Go Testing
 *************************/
func main() {

	// 2K / 4K / 8K / 16K / 32K resolutions
	Resolution := 0

	var Julia_Const = complex(-0.4, 0.6)
	var Max_Iterations = 255
	var X_Dim [5]int = [5]int{2560, 3840, 7680, 15360, 30720}
	var Y_Dim [5]int = [5]int{1440, 2160, 4320, 8640, 17280}
	Image_Width := 0
	Image_Height := 0

	// --------------------------------------
	// ------ Select Image Dimensions -------
	Image_Width = X_Dim[Resolution];
	Image_Height = Y_Dim[Resolution];


	// ------------------------------------
	// ------ Printout Configuration ------
	fmt.Printf("Julia Constant  : [ %[1]f %[2]f i] \n", real(Julia_Const), imag(Julia_Const))
	fmt.Printf("Image Resolution: [ %[1]d x %[2]d ]\n", Image_Width, Image_Height)
	fmt.Printf("Canvas Area    X: [ %[1]f --> %[2]f ]\n", X_Min, X_Max)
	fmt.Printf("Canvas Area    Y: [ %[1]f --> %[2]f ]\n", Y_Min, Y_Max)

	// -------------------------------------
	// ------- Detect Amount of CPUs -------
	var NumCPUs int = runtime.NumCPU()
	fmt.Println("Amount of CPUs Detected: ", NumCPUs)

	// -------------------------------------
	// ----- Allocate Buffer for Image -----
	Fractal_Image := image.NewNRGBA(image.Rect(0, 0, Image_Width, Image_Height))

	// -------------------------------------------------
	// ------- Runs some "Multitreaded Routines" -------
	var WaitGroup sync.WaitGroup
	var NumThreads = NumCPUs
	y_chunk := Image_Height / NumThreads
	start := time.Now()
	for i := 0; i < NumThreads; i++ {
		WaitGroup.Add(1)
		x_start := 0
		x_stop := Image_Width
		y_start := i * y_chunk
		y_stop := (i + 1) * y_chunk
		Console_Line := fmt.Sprintf("Chunk [ %2[1]d ]: X[ %[2]d - %4[3]d ] Y[ %4[4]d - %4[5]d ]", i, x_start, x_stop, y_start, y_stop)
		fmt.Println(Console_Line)

		// -------------------------
		// ------ Get Going! -------
		go julia_render(&WaitGroup, Fractal_Image, Image_Width, Image_Height, x_start, x_stop, y_start, y_stop, Julia_Const, Max_Iterations, i)

	}

	// ----------------------------------------------
	// ------ Wait for All Subroutines to Exit ------
	WaitGroup.Wait()

	// -------------------------------------
	// ------- Printout Elapsed Time -------
	Elapsed_Time := time.Since(start)
	fmt.Println("Elapsed Time: ",Elapsed_Time);

	// --------------------------------------------
	// ------ Encode as PNG and save to Disk ------
	f, _ := os.Create("Julia.png")
	png.Encode(f, Fractal_Image)


	// ------------------------------------------------------
	// ------ Give enough Time to see Calculation Time ------
	duration := time.Duration(5)*time.Second
	time.Sleep(duration)

}
