/********************
 * IrisCores.com 2023
 ********************/

// ---------------------------------------------------
// ------- Ignore some standard rust's bullshit ------ 
#![allow(unused_variables)]
#![allow(unused_parens)]
#![allow(unused_labels)]
#![allow(unused_assignments)]
#![allow(dead_code)]
#![allow(non_snake_case)]

use num::Complex;
use std::time::*;
use std::fs::File;
use image::ColorType;
use rgb::RGB8;
use image::codecs::png::PngEncoder;
use image::ImageEncoder;


/********************
 * Canvas Definitions 
 ********************/
const X_MIN : f64 = -2.0;
const X_MAX : f64 = 2.0;
const Y_MIN : f64 = -2.0;
const Y_MAX : f64 = 2.0; 
const CL3 : usize = 3; // RGB 
const CL4 : usize = 4; // RGBA

/**************************************
 * @brief: Main Julia Computing Routine 
 **************************************/
fn julia_compute(Plane_Point : Complex<f64>, Julia: Complex<f64>, maxiter: usize) -> (usize) {
    let mut z = Plane_Point;
    for i in 0..maxiter {
        let abs_val = z.norm_sqr();
        if abs_val > 4.0 {
            return i;
        }
        z = z * z + Julia;
    }
  
    return 0;
}


/*********************************************************************************************
 *@brief : Main thread-wise Julia Rendering
 * WARNING: Each Thread have a slice of the original buffer (IE Not Entire Buffer!)
 * The result is that the indexing cannot start from 0 and end up at Image_Width * Image_Heigh
 * Instead,each thread is given its offset to compensate for the slicing.
 **********************************************************************************************/
fn julia_render(pixels: &mut [RGB8],x_dim:usize,y_dim:usize,x_start :usize,x_stop :usize,y_start:usize,y_stop :usize,offset :usize){

    let X_Rend:f64 = X_MAX - X_MIN;
	  let Y_Rend:f64 = Y_MAX - Y_MIN;

    // --------------------------
    // ------- Get ColorMap ------
    let cl_map = colorous::MAGMA;

    // ----------------------------------------------------
    // ------- Compute the Dedicated Piece of Canvas ------
    'y_dim_loop: for y in y_start..y_stop {
        'x_dim_loop:for x in x_start..x_stop {

            // -----------------------------------------------
            // ------ Compute Point in 2D Complex Plane ------
            let Point_Real:f64 = X_MIN + (x as f64)/(x_dim as f64)*X_Rend;
            let Point_Imag:f64 = Y_MIN + (y as f64)/(y_dim as f64)*Y_Rend;
            let Plane_Point = Complex::new(Point_Real,Point_Imag);

            // -----------------------------------
            // ------ Julia Computing Value ------
            let Julia_Value = Complex::new(-0.4,0.6);

            let Max_Iterations = 255;
            let mut Num_Iterations = julia_compute(Plane_Point, Julia_Value, Max_Iterations);
            if Num_Iterations > 0 {Num_Iterations = Max_Iterations - Num_Iterations;}
  
            // ---------------------------------
            // ------ Convert ColorSpace -------
            let Normalized_Value = Num_Iterations as f64 / 255.0;
            let Pixel_Rgb = cl_map.eval_continuous(Normalized_Value); 


            // ------------------------------
            // ------- Write to Buffer ------
            let index = y * x_dim + x;
            pixels[index - offset].r = Pixel_Rgb.r as u8;
            pixels[index - offset].g = Pixel_Rgb.g as u8;
            pixels[index - offset].b = Pixel_Rgb.b as u8;

        }
    }
}

/*******************************************
 *@brief: Write the rendered canvas into png
 ******************************************/
fn write_image(filename: &str, pixels: &[RGB8], image_width: usize,image_heigh: usize){

    // --------------------------------
    // ------ Create File Handle ------
    let output = File::create(filename).unwrap();

    // ----------------------------
    // ------ Create Encoder ------
    let encoder = PngEncoder::new(output);

    // ----------------------------------------------
    // ------- Convert to Different Data Type -------
    let mut pixels_u8: Vec<u8> = vec![0; CL4 * image_width * image_heigh];
    'row_loop:for y in 0..image_heigh {
        'column_loop:for x in 0..image_width {
            let index = y * image_width + x;
            pixels_u8[index * CL4 + 0] = pixels[index].r;
            pixels_u8[index * CL4 + 1] = pixels[index].g;
            pixels_u8[index * CL4 + 2] = pixels[index].b;
            pixels_u8[index * CL4 + 3] = 255; // Alpha Channel
        }
    } 

    // --------------------------------------
    // ------ Convert and Save to File ------
    encoder.write_image(&pixels_u8,image_width as u32, image_heigh as u32,ColorType::Rgba8).unwrap();

}

fn main() {
    let Num_Threads = 20;
    let Resolution = 3;
    let X_Resolutions = [2560,3840,7680,15360];
    let Y_Resolutions = [1440,2160,4320,8640];
    let Image_Width = X_Resolutions[Resolution];
    let Image_Heigh = Y_Resolutions[Resolution];
    let Num_Pixels = Image_Width * Image_Heigh;
    let Default_Pixel = RGB8{r:0,g:0,b:0};

    println!("Num Threads: {:?}",Num_Threads);
    println!("Image Dimensions: {:?}",(Image_Width,Image_Heigh));

    // ----------------------------------------------
    // ------- Allocate Memory For Computation ------
    let mut Julia_Raw:Vec<RGB8> = vec![Default_Pixel;Num_Pixels];
    let Sliced_Julia:&mut[RGB8] = Julia_Raw.as_mut_slice();

    // ----------------------------------------
    // ------ Slice Computing Into Parts ------
    let iter = Sliced_Julia.chunks_mut(Num_Pixels / Num_Threads);
    let start = Instant::now();
    rayon::scope(|s| {
        
        // ------------------------------------
        // ------ Loop Across all Chunks ------
        for(i,slice) in iter.enumerate(){
          let x_start  = 0;
          let x_stop   = Image_Width;
          let y_start  = i * (Image_Heigh / Num_Threads);
          let y_stop   = (i + 1) * (Image_Heigh / Num_Threads);
          let offset   = i * (Image_Heigh / Num_Threads) * Image_Width;
          
          // ----------------------------------------------------
          // ------ Spawn Thread on the Rayon's Threadpool ------
          s.spawn(move|_| julia_render(slice,Image_Width,Image_Heigh,x_start,x_stop,y_start,y_stop,offset));
        }
      
    });

    let duration = start.elapsed();
    println!("Julia Computation Finished in: {:?}", duration);

    // ----------------------------------------
    // ------ Write Final Image to File -------
    write_image(String::from("Julia.png").as_str(), &Julia_Raw, Image_Width,Image_Heigh);
}
